import json
from http import HTTPStatus
from dataclasses import dataclass, InitVar, asdict
from typing import List, Iterable, TypeVar


HTTPCode = TypeVar('HTTPCode', HTTPStatus, int)


@dataclass(frozen=True)
class ConstraintError(Exception):
    message: str


@dataclass(frozen=True)
class HTTPConstraintError(ConstraintError):
    status: InitVar[HTTPCode]

    def __post_init__(self, status):
        super().__setattr__('status', HTTPStatus(status))


class ConstraintsErrors(Exception):
    errors: List[ConstraintError]

    def __init__(self, *errors: ConstraintError):
        self.errors = list(errors)

    def __iter__(self):
        return iter(self.errors)

    def __len__(self):
        return len(self.errors)

    def __eq__(self, other):
        if isinstance(other, ConstraintsErrors):
            return self.errors == other.errors
        elif isinstance(other, Iterable):
            return self.errors == other
        return False

    def json(self):
        return json.dumps([asdict(e) for e in self.errors])



import enum
from abc import ABC, abstractproperty
from collections import defaultdict
from typing import Type, Optional, Dict, Callable
from roughrider.workflow.components import State, Transitions, Transition


class WorkflowState(State, enum.Enum):
    pass


class WorkflowItem(ABC):

    state: WorkflowState

    def __init__(self, workflow, item, **namespace):
        self.workflow = workflow
        self.item = item
        self.namespace = namespace

    @abstractproperty
    def state(self):
        pass

    def get_possible_transitions(self):
        return tuple(self.workflow.transitions.available(
             self.state, self.item, **self.namespace))

    def get_transition(self, target: State) -> Optional[Transition]:
        target = self.workflow.states(target)  # idempotent
        return self.workflow.transitions.find(self.state, target)

    def apply_transition(self, transition: Transition):
        error = transition.action.check_constraints(
            self.item, **self.namespace)
        if error is not None:
            raise error
        self.state = transition.target

    def transition_to(self, state: State):
        transition = self.get_transition(state)
        self.apply_transition(transition)


class Workflow:

    states: Type[WorkflowState]
    wrapper: Type[WorkflowItem]
    transitions: Transitions
    default_state: WorkflowState

    def __init__(self, default_state):
        if default_state in self.states.__members__:
            self.default_state = self.states[default_state]
        else:
            self.default_state = self.states(default_state)

    def __getitem__(self, name: str) -> WorkflowState:
        return self.states[name]

    def __call__(self, item, **namespace):
        return self.wrapper(self, item, **namespace)

    def get(self, name=None):
        if name is None:
            return self.default_state
        return self.states[name]



import pytest
from roughrider.workflow.components import Action, Transition, Transitions
from roughrider.workflow.workflow import WorkflowState
from roughrider.predicate.validators import ConstraintError


def only_test(item: str):
    if item != 'test':
        raise ConstraintError('We need a test')


class states(WorkflowState):
    foo = 'Foo'
    bar = 'Bar'


forward = Transition(
    origin=states.foo,
    target=states.bar,
    action=Action('Foobaring')
)

backward = Transition(
    origin=states.bar,
    target=states.foo,
    action=Action(
        'Barfooing',
        constraints=[only_test],
    )
)


def test_transition():
    assert forward.origin == states.foo
    assert forward.target == states.bar
    assert forward.action == Action('Foobaring')

    assert backward.origin == states.bar
    assert backward.target == states.foo
    assert backward.action == Action(
        'Barfooing',
        constraints=[only_test],
    )


def test_init_transition_faulty():

    with pytest.raises(TypeError) as exc:
        Transition()

    assert str(exc.value) == (
        "__new__() missing 3 required positional arguments: "
        "'action', 'origin', and 'target'"
    )


def test_transitions_internal_edges():

    transitions = Transitions((forward,))

    assert isinstance(transitions, tuple)
    assert isinstance(transitions._edges, dict)
    assert dict(transitions._edges) == {
        states.foo: {
            states.bar: forward
        }
    }

    transitions = Transitions((forward, backward))
    assert dict(transitions._edges) == {
        states.foo: {
            states.bar: forward
        },
        states.bar: {
            states.foo: backward
        }
    }


def test_transitions_find():

    transitions = Transitions((forward,))

    assert transitions.find(states.foo, states.bar) == forward

    with pytest.raises(LookupError) as exc:
        transitions.find(states.bar, states.foo)

    assert str(exc.value) == 'No transition from states.bar to states.foo'

    transitions = Transitions((forward, backward))
    assert transitions.find(states.foo, states.bar) == forward
    assert transitions.find(states.bar, states.foo) == backward


def test_transitions_available():

    transitions = Transitions((forward, backward))
    assert list(transitions.available(states.foo, 'test')) == [forward]
    assert list(transitions.available(states.bar, 'test')) == [backward]
    assert list(transitions.available(states.bar, 'nope')) == []



from roughrider.workflow.components import Action
from roughrider.predicate import Validator, ConstraintError, ConstraintsErrors


def test_no_constraints_action():
    action = Action(identifier='Test')
    assert action.identifier == 'Test'
    assert action.constraints == []
    assert not action.check_constraints('item')


def test_constraint_action():

    def tester(item, **namespace):
        if item != 'test':
            raise ConstraintError('This needs to be a test')

    action = Action(identifier='Test', constraints=[tester])
    assert action.identifier == 'Test'
    assert action.constraints == [tester]

    errors = action.check_constraints('not test')
    assert isinstance(errors, ConstraintsErrors)
    assert len(errors) == 1
    assert list(errors) == [ConstraintError('This needs to be a test')]

    assert not action.check_constraints('test')



from abc import ABC, abstractmethod
from typing import Iterable, Tuple, Optional, Callable, Union
from roughrider.predicate.errors import ConstraintError, ConstraintsErrors


class Validator(ABC):
    """A validator.
    """
    description: Optional[str]

    @abstractmethod
    def __call__(self, *args, **namespace) -> None:
        """Raises a roughrider.predicate.ConstraintError if the validation failed.
        """


Constraint = Union[Validator, Callable[..., None]]


class Or(Tuple[Constraint], Validator):

    def __call__(self, *args, **namespace):
        errors = []
        for validator in self:
            try:
                validator(*args, **namespace)
                return
            except ConstraintError as exc:
                errors.append(exc)
            except ConstraintsErrors as exc:
                errors.extend(exc.errors)

        raise ConstraintsErrors(*errors)


def resolve_validators(
        validators: Iterable[Constraint], *args, **namespace
) -> Optional[ConstraintsErrors]:
    errors = []
    for validator in validators:
        try:
            validator(*args, **namespace)
        except ConstraintError as exc:
            errors.append(exc)
        except ConstraintsErrors as exc:
            errors.extends(exc.errors)
    if errors:
        return ConstraintsErrors(*errors)



from collections import defaultdict
from typing import NamedTuple, Iterable, Tuple, Optional, Mapping
from roughrider.predicate.validators import Constraint, resolve_validators
from roughrider.predicate.errors import ConstraintsErrors


class State:
    identifier: str

    def __init__(self, identifier: str):
        self.identifier = identifier

    def __hash__(self):
        return hash(self.identifier)


class Action:

    identifier: str
    constraints: Iterable[Constraint]

    def __init__(self, identifier: str, constraints: Optional[list] = None):
        self.identifier = identifier
        self.constraints = constraints if constraints is not None else []

    def __eq__(self, other):
        if isinstance(other, Action):
            return (self.identifier == other.identifier and
                    self.constraints == other.constraints)
        return False

    def check_constraints(self, item, **ns) -> Optional[ConstraintsErrors]:
        """Checks the constraints against the given object.
        """
        if self.constraints:
            return resolve_validators(self.constraints, item, **ns)


class Transition(NamedTuple):
    action: Action
    origin: State
    target: State


class Transitions(Tuple[Transition]):

    _edges: Mapping[State, Mapping[State, Action]] = None

    def __new__(cls, transitions: Iterable[Transition]):
        obj = super().__new__(Transitions, transitions)
        obj._edges = defaultdict(dict)
        for trn in transitions:
            obj._edges[trn.origin][trn.target] = trn
        return obj

    def available(self, origin, item, **ns):
        for target, trn in self._edges[origin].items():
            if trn.action.check_constraints(item, **ns) is None:
                yield trn

    def find(self, origin, target):
        try:
            return self._edges[origin][target]
        except KeyError:
            raise LookupError(f'No transition from {origin} to {target}')



import pytest
from roughrider.predicate import ConstraintError, Validator, Or
from roughrider.workflow.components import Action, Transition, Transitions
from roughrider.workflow.workflow import (
    WorkflowItem, WorkflowState, Workflow)


class Document:
    state = None
    body = ""


def NonEmptyDocument(item, **namespace):
    if not item.body:
        raise ConstraintError(message='Body is empty.')


class RoleValidator(Validator):

    def __init__(self, role):
        self.role = role

    def __call__(self, item, role=None, **namespace):
        if role != self.role:
            raise ConstraintError(
                message=f'Unauthorized. Missing the `{role}` role.')


class PublicationWorkflow(Workflow):

    class wrapper(WorkflowItem):

        @property
        def state(self):
            return self.workflow.get(self.item.state)

        @state.setter
        def state(self, state):
            self.item.state = state.name


    class states(WorkflowState):
        draft = 'Draft'
        published = 'Published'
        submitted = 'Submitted'


    transitions = Transitions((
        Transition(
            origin=states.draft,
            target=states.published,
            action=Action(
                'Publish',
                constraints=[NonEmptyDocument, RoleValidator('publisher')]
            )
        ),
        Transition(
            origin=states.published,
            target=states.draft,
            action=Action(
                'Retract',
                constraints=[
                    Or((RoleValidator('owner'),
                        RoleValidator('publisher')))
                ]
            )
        ),
        Transition(
            origin=states.draft,
            target=states.submitted,
            action=Action(
                'Submit',
                constraints=[NonEmptyDocument, RoleValidator('owner')],
            )
        ),
        Transition(
            origin=states.submitted,
            target=states.published,
            action=Action(
                'Publish',
                constraints=[NonEmptyDocument, RoleValidator('publisher')],
            )
        )
    ))


workflow = PublicationWorkflow('draft')


def test_publish_worflow():
    item = Document()
    workflow_item = workflow(item, role='some role')
    assert workflow_item.state == workflow.get('draft')
    assert not workflow_item.get_possible_transitions()

    item.body = "Some text here"
    assert not workflow_item.get_possible_transitions()

    workflow_item = workflow(item, role='owner')
    assert workflow_item.get_possible_transitions() == (
        workflow.transitions[2],
    )

    workflow_item.transition_to(PublicationWorkflow.states.submitted)
    assert workflow_item.state == workflow.get('submitted')



from functools import wraps
from typing import Callable, Any
from roughrider.predicate.errors import ConstraintError


PredicateErrorHandler = Callable[[ConstraintError], Any]


def with_predicates(predicates, handler=None):
    def predication_wrapper(func):
        @wraps(func)
        def assert_predicates(*args, **kwargs):
            for predicate in predicates:
                try:
                    predicate(*args, **kwargs)
                except ConstraintError as exc:
                    if handler is not None:
                        return handler(exc)
                    raise
            return func(*args, **kwargs)
        return assert_predicates
    return predication_wrapper



from roughrider.workflow import WorkflowState, State


def test_state():

    state = State('test')
    assert state.identifier == 'test'

    state = State(identifier='test')
    assert state.identifier == 'test'

    assert hash(state) == hash('test')


def test_workflow_states():

    class states(WorkflowState):
        foo = 'Foo'
        bar = 'Bar'

    assert isinstance(states.foo, State)
    assert isinstance(states.foo, WorkflowState)
    assert states.foo.name == 'foo'
    assert states.foo.identifier == 'Foo'

    assert states('Foo') is states.foo
    assert states(states.foo) is states.foo
    assert states['foo'] is states.foo



from .components import Action, State, Transition, Transitions
from .workflow import Workflow, WorkflowState, WorkflowItem


__all__ = [
    'Action',
    'State',
    'Transition',
    'Transitions',
    'Workflow',
    'WorkflowItem',
    'WorkflowState',
]



from roughrider.predicate import ConstraintError, Validator, Or
from roughrider.workflow.components import Action, Transition, Transitions
from roughrider.workflow.workflow import (
    WorkflowItem, WorkflowState, Workflow)


class Document:
    state = None
    body = ""


class RoleValidator(Validator):

    def __init__(self, role):
        self.role = role

    def __call__(self, item, role=None, **namespace):
        if role != self.role:
            raise ConstraintError(
                message=f'Unauthorized. Missing the `{role}` role.')


class PublicationWorkflow(Workflow):

    class wrapper(WorkflowItem):

        @property
        def state(self):
            return self.workflow.get(self.item.state)

        @state.setter
        def state(self, state):
            self.item.state = state.name


    class states(WorkflowState):
        draft = 'Draft'
        published = 'Published'
        submitted = 'Submitted'


    transitions = Transitions((
        Transition(
            origin=states.draft,
            target=states.published,
            action=Action(
                'Publish',
                constraints=[RoleValidator('publisher')]
            )
        ),
        Transition(
            origin=states.published,
            target=states.draft,
            action=Action(
                'Retract',
                constraints=[
                    Or((RoleValidator('owner'),
                        RoleValidator('publisher')))
                ]
            )
        ),
        Transition(
            origin=states.draft,
            target=states.submitted,
            action=Action(
                'Submit',
                constraints=[RoleValidator('owner')],
            )
        ),
        Transition(
            origin=states.submitted,
            target=states.published,
            action=Action(
                'Publish',
                constraints=[RoleValidator('publisher')],
            )
        )
    ))


workflow = PublicationWorkflow('draft')  # initial state
item = Document()
workflow_item = workflow(item, role='owner')
workflow_item.transition_to(PublicationWorkflow.states.submitted)



from .validators import Or, Validator, resolve_validators
from .errors import ConstraintError, HTTPConstraintError, ConstraintsErrors


__all__ = [
    'ConstraintsErrors',
    'Error',
    'HTTPConstraintError',
    'Or',
    'Validator',
    'resolve_validators',
]
